<?php
namespace SupplyTek\Controllers;

use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use SupplyTek\Utilities\CrudSupplyTek;
use SupplyTek\Utilities\SessionHandler;
use SupplyTek\Utilities\FormatData;
use SupplyTek\Utilities\TemplateHandler;
use SupplyTek\Utilities\EmailHandler;
use SupplyTek\Utilities\JWT;
use SupplyTek\Services\LoginService;

/**
 * Handles Login and Generate Token
 * 
 * Controls login requests, retrieving user data from database and
 * verifying if the user exists. Generates a JWT Token for validation
 * and stores data from user in PHP Session
 *
 * @author Rômulo Alves Lousada
 */
class LoginController 
{
    /**
     * Timer for Token Expiration
     *
     * Predefined timer to add seconds to token expiration time.
     * Expected values are:
     * 
     * 43200 - 12 Hours
     * 86400 - 1 Day
     * 172800 - 2 Days
     * 604800 - 1 Week
     *
     * @var int
     */
    const TOKEN_TIMER = 86400;

    protected $crud;
    protected $session;
    protected $format;
    protected $template;
    protected $email;
    protected $jwt;
    protected $service;

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param CrudSupplyTek $crudSupplyTek Object that allows to perform queries from datasource
     * @param SessionHandler $sessionHandler Object that allows to manipulate session
     * @param FormatData $formatData Object that allows to format data
     * @param TemplateHandler $templateHandler Object that allows to retrive templates
     * @param EmailHandler $emailHandler Object that allows to send mails
     * @param JWT $jwt Object that allows to create and validate tokens
     * @param LoginService $loginService Object to access LoginService
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function __construct (CrudSupplyTek $crudSupplyTek, SessionHandler $sessionHandler, FormatData $formatData, TemplateHandler $templateHandler, EmailHandler $emailHandler, JWT $jwt, LoginService $loginService) 
    {
        $this->crud = $crudSupplyTek;
        $this->session = $sessionHandler;
        $this->format = $formatData;
        $this->template = $templateHandler;
        $this->email = $emailHandler;
        $this->jwt = $jwt;
        $this->service = $loginService;
        
        $this->sessId = 0;
    }

    /**
     * Receives and validade information from login
     * 
     * Receives the data sent from the frontend and calls functions from the 
     * providers to search the database for a registry.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function login(Request $request, Response $response)
    {
        try {
            $rows = $request->getDataSet()->getRows();
            
            if (isset($rows[0]["email"]) && isset($rows[0]["password"])) {
                $params = array(
                    "EMAIL" => $rows[0]["email"],
                    "PASSWORD" => md5($rows[0]["password"])
                ); 

                $user = $this->crud->executeQuery("auth", $params);
            } else {
                throw new \Exception("Favor preencher os campos do formulário e tentar novamente.", 400);
            }
            
            if (isset($user[0])) {
                $this->session->clearSession();
                $this->storeDataOnSession($user[0], "user_data");
                
                $paramGeral = $this->crud->executeQuery("paramgeral-data");
                $return = array();
                
                if (isset($paramGeral[0])) {
                    $this->storeDataOnSession($paramGeral[0], "param_geral_data");
                    $return["mask"] = $paramGeral[0]["CDPICTPROD"];
                }
                
                $return["token"] = $this->generateToken($user[0]);
                $return["changePassword"] = $this->verifyChangePassword($user[0]);
                
                $response->setMetadata($return);
            } else {
                throw new \Exception("Usuário ou senha inválidos.", 401);
            }
        } catch (\Exception $e) {
            throw new \Exception ($e->getMessage(), $e->getCode());
        }
    }

    /**
     * Store data on the session.
     * 
     * Stores data on the session. 
     * These are sensible data that are returned during the login process.
     * 
     * @param array $data Array with data to be saved.
     * @param string $path String with the path where the data will be saved.
     *
     * @return void
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function storeDataOnSession($data, $path)
    {
        foreach ($data as $index => $value) {
            $this->session->setSessionData($index, $value, $path);
        }
    }

    /**
     * Token Generator
     * 
     * Generates a Token based on the login data passed and return it.
     *
     * @param array $user – Array with logged user data.
     *
     * @return string
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function generateToken($user)
    {
        $arrayDataToken = array(
            "operador" => $user["CDOPERADOR"],
            "org" => $user["NRORG"],
            "exp" => time() + self::TOKEN_TIMER
        );

        return $this->jwt->generateToken($arrayDataToken);
    }
    
    /**
     * Verify Password Changed
     * 
     * Checks if password has been changed and returns a bool depending
     * on the result of the query.
     *
     * @param array $user – Array with logged user data.
     *
     * @return bool
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function verifyChangePassword($user)
    {
        $params["OPERATOR_ID"] = $user["CDOPERADOR"];
        
        $changePassword = $this->crud->executeQuery("verify-change-password", $params);
        
        if (isset($changePassword[0])) {
            return $changePassword[0]["IDATUCTRLSENHA"] == "R" ? true : false;
        }
        
        return false;
    }
    
    /**
     * Generate new password
     * 
     * Receives an email, checks on the database if the email is from a valid user
     * to access SupplyTek. If it's valid, generate a new password that will be
     * encrypted and saved on the table, and then sent to the email used on the
     * process.
     *
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return void
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function forgotPassword(Request $request, Response $response)
    {
        try {
            $this->service->beginTransaction();
            $rows = $request->getDataSet()->getRows();

            if (isset($rows[0]["formData"]["email"])) {
                $params = array(
                    "EMAIL" => $rows[0]["formData"]["email"]
                ); 

                $user = $this->crud->executeQuery("send-email-new-password", $params);
            } else {
                throw new \Exception("Favor preencher os campos do formulário e tentar novamente.", 400);
            }

            if (isset($user[0])) {
                $this->storeDataOnSession($user[0], "user_data");
    
                $password = $this->format->generateRandomString();
                
                $this->getAdditionalParameters($params, "R");
                $this->encryptPasswords($params, $password);
                
                $this->service->insertControl($params);
                $this->service->updatePassword($params);
                
                $email = $this->createEmailContent($user[0], $params["EMAIL"], $password);
                $this->email->useCustomSmtp();
                $this->email->send($email["SUBJECT"], $email["CONTENT"], $params["EMAIL"]);
                $this->service->commitTransaction();
            } else {
                throw new \Exception("Email não cadastrado no sistema. Favor verificar.", 401);
            }
        } catch (\Exception $e) {
            $this->service->rollbackTransaction();
            throw new \Exception ($e->getMessage(), $e->getCode());
        }
    }
    
    /**
     * Insert additional parameters on array 
     * 
     * Receives an array by reference that will have additional
     * data created on it. This data is needed on the insert/update process
     * of forgot password method.
     *
     * @param array $params Array that will have data added to it.
     * @param string $type (optional) Type of the update: R or S
     *
     * @return void
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function getAdditionalParameters(&$params, $type = "")
    {
        $params["ORG_ID"] = $this->session->getSessionData("NRORG");
        $params["OPERATOR_ID"] = $this->session->getSessionData("CDOPERADOR");
        
        if ($type != "") {
            $params["UPDATE_CTRLSENHA_ID"] = $type;
        }
    }

    /**
     * Password encryption
     * 
     * Encrypts the password to web(md5) and delphi(ISO-8859-1)
     * formats and save them on the array that will return with
     * the data to be used on insert and update.
     *
     * @param array $params Array that will have data added to it.
     * @param string $password Random string containing the generated password.
     *
     * @return void
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function encryptPasswords(&$params, $password)
    {
        $params["PASSWORD_DELPHI"] = $this->email->getDelphiPasswordEncrypted($password);
        $params["PASSWORD_WEB"] = md5($password);
    }
    
    /**
     * Create all text content of the mail
     * 
     * Creates and returns an array with texts needed to be sent
     * via mail to the user.
     *
     * @param array $user Array with user data returned from database
     * @param string $mail String containing the informed mail on the forgot password screen.
     * @param string $password Random string containing the generated password.
     *
     * @return array
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function createEmailContent($user, $mail, $password) 
    {
        $file = array("name" => "forgot-password", "extension" => ".html");
        $replace = array(
            "%OPERATOR_NAME%" => $this->format->capitalizeWords($user["NMOPERADOR"]), 
            "%OPERATOR_EMAIL%" => $mail, 
            "%OPERATOR_PASSWORD%" => $password
        );

        $email = array();
        
        $email["SUBJECT"] = "Alteração de Senha";
        
        $email["CONTENT"] = $this->template->getTemplate($file, $replace);
        
        return $email;
    }
    
    /**
     * Save new password
     * 
     * Verifies if all the fields are filled, and then process the passwords
     * to check if the information provided is valid. If it pass on all
     * checks, updates the database with the new password.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return void
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function saveNewPassword(Request $request, Response $response)
    {
        try {
            $this->service->beginTransaction();
            $rows = $request->getDataSet()->getRows();

            if (
                isset($rows[0]["formData"]["actual-password"]) &&
                isset($rows[0]["formData"]["new-password"]) &&
                isset($rows[0]["formData"]["new-password-confirm"])
            ) {
                $actualPassword = $rows[0]["formData"]["actual-password"];
                $newPassword = $rows[0]["formData"]["new-password"];
                $newPasswordConfirmation = $rows[0]["formData"]["new-password-confirm"];
                
                $this->getAdditionalParameters($params);
                $this->encryptPasswords($params, $actualPassword);

                $check = $this->crud->executeQuery("check-actual-password", $params);
                if (isset($check[0]["CHECK"]) && $check[0]["CHECK"]) {
                    $check = $this->validateNewPassword($newPassword, $newPasswordConfirmation);
                    
                    if ($check) {
                        $params["UPDATE_CTRLSENHA_ID"] = "S";
                        $this->encryptPasswords($params, $newPassword);
                        $this->service->insertControl($params);
                        $this->service->updatePassword($params);
                        $this->service->commitTransaction();
                    } 
                } else {
                    throw new \Exception("A senha atual fornecida está incorreta.", 401);
                }
            } else {
                throw new \Exception("Favor preencher os campos do formulário e tentar novamente.", 400);
            }
        } catch (\Exception $e) {
            $this->service->rollbackTransaction();
            throw new \Exception ($e->getMessage(), $e->getCode());
        }
    }
    
    /**
     * Validate new password
     * 
     * Receives the new password and it's confirmation, then validate if they are equal and
     * if the new password is according to the validation:
     * 
     * At least 2 uppercase letters;
     * At least 3 lowercase letters;
     * At least 2 numbers;
     * At least 1 special character;
     * String must be between 8 and 12 characters
     *
     * @param string $newPassword String containing new password.
     * @param string $newPasswordConfirmation String containing new password confirmation.
     *
     * @return bool
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function validateNewPassword($newPassword, $newPasswordConfirmation)
    {
        if ($newPassword === $newPasswordConfirmation) {
            $uppercase = preg_match_all('@[A-Z]@', $newPassword);
            $lowercase = preg_match_all('@[a-z]@', $newPassword);
            $number = preg_match_all('@[0-9]@', $newPassword);
            $specialChars = preg_match_all('@[^\w]@', $newPassword);
            
            if ($lowercase >= 3 && $uppercase >= 2 && $number >= 2 && $specialChars >= 1 && (strlen($newPassword) >= 8 && strlen($newPassword) <= 12)) {
                return true;
            } else {
                throw new \Exception("A nova senha deve possuir no mínimo: 3 caracteres minusculos, 2 caracteres maiúsculos, 1 caractere especial, 1 número e entre 8 a 12 caracteres.", 400);
            }
        } else {
            throw new \Exception("A nova senha é diferente da confirmação", 400);
        }
    }
    
    /**
     * Do the logout from the current user logged.
     * 
     * Clears the session data stored, and returns a confirmation to front
     * that will proceed to send the user to the login page.
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function logout(Request $request, Response $response)
    {
        try {
            $this->session->clearSession();
            
            $response->setMetadata(["OK"]);
        } catch (\Exception $e) {
            throw new \Exception ($e->getMessage(), $e->getCode());
        }
    }
}